@@ -25,16 +25,23 @@ import (
 	"os"
 	"sort"
 	"strings"
+	"sync"
 	"testing"
 	"time"
 
 	"github.com/cespare/xxhash"
 	"github.com/dgraph-io/badger/v2/options"
+	"github.com/dgraph-io/badger/v2/pb"
 	"github.com/dgraph-io/badger/v2/y"
 	"github.com/dgraph-io/ristretto"
 	"github.com/stretchr/testify/require"
 )
 
+const (
+	KB = 1024
+	MB = KB * 1024
+)
+
 func key(prefix string, i int) string {
 	return prefix + fmt.Sprintf("%04d", i)
 }
@@ -351,7 +358,7 @@ func TestIterateBackAndForth(t *testing.T) {
 
 	it.seekToFirst()
 	k = it.Key()
-	require.EqualValues(t, key("key", 0), string(y.ParseKey(k)))
+	require.EqualValues(t, key("key", 0), y.ParseKey(k))
 }
 
 func TestUniIterator(t *testing.T) {
@@ -696,8 +703,7 @@ func TestTableBigValues(t *testing.T) {
 	require.NoError(t, err, "unable to create file")
 
 	n := 100 // Insert 100 keys.
-	opts := Options{Compression: options.ZSTD, BlockSize: 4 * 1024, BloomFalsePositive: 0.01,
-		TableSize: uint64(n) * 1 << 20}
+	opts := Options{Compression: options.ZSTD, BlockSize: 4 * 1024, BloomFalsePositive: 0.01}
 	builder := NewTableBuilder(opts)
 	for i := 0; i < n; i++ {
 		key := y.KeyWithTs([]byte(key("", i)), 0)
@@ -782,7 +788,7 @@ func BenchmarkReadAndBuild(b *testing.B) {
 	for i := 0; i < b.N; i++ {
 		func() {
 			opts := Options{Compression: options.ZSTD, BlockSize: 4 * 0124, BloomFalsePositive: 0.01}
-			opts.Cache = cache
+			opts.BlockCache = cache
 			newBuilder := NewTableBuilder(opts)
 			it := tbl.NewIterator(false)
 			defer it.Close()
@@ -808,7 +814,7 @@ func BenchmarkReadMerged(b *testing.B) {
 	for i := 0; i < m; i++ {
 		filename := fmt.Sprintf("%s%s%d.sst", os.TempDir(), string(os.PathSeparator), rand.Int63())
 		opts := Options{Compression: options.ZSTD, BlockSize: 4 * 1024, BloomFalsePositive: 0.01}
-		opts.Cache = cache
+		opts.BlockCache = cache
 		builder := NewTableBuilder(opts)
 		f, err := y.OpenSyncedFile(filename, true)
 		y.Check(err)
@@ -900,7 +906,7 @@ func getTableForBenchmarks(b *testing.B, count int, cache *ristretto.Cache) *Tab
 		cache, err = ristretto.NewCache(&cacheConfig)
 		require.NoError(b, err)
 	}
-	opts.Cache = cache
+	opts.BlockCache = cache
 	builder := NewTableBuilder(opts)
 	filename := fmt.Sprintf("%s%s%d.sst", os.TempDir(), string(os.PathSeparator), rand.Int63())
 	f, err := y.OpenSyncedFile(filename, true)
@@ -932,3 +938,41 @@ func TestOpenKVSize(t *testing.T) {
 	var entrySize uint64 = 15 /* DiffKey len */ + 4 /* Header Size */ + 4 /* Encoded vp */
 	require.Equal(t, entrySize, table.EstimatedSize())
 }
+
+// Run this test with command "go test -race -run TestDoesNotHaveRace"
+func TestDoesNotHaveRace(t *testing.T) {
+	opts := getTestTableOptions()
+	f := buildTestTable(t, "key", 10000, opts)
+	table, err := OpenTable(f, opts)
+	require.NoError(t, err)
+	defer table.DecrRef()
+
+	var wg sync.WaitGroup
+	wg.Add(5)
+	for i := 0; i < 5; i++ {
+		go func() {
+			require.True(t, table.DoesNotHave(uint64(1237882)))
+			wg.Done()
+		}()
+	}
+	wg.Wait()
+}
+
+var ko *pb.BlockOffset
+
+// Use this benchmark to manually verify block offset size calculation
+func BenchmarkBlockOffsetSizeCalculation(b *testing.B) {
+	for i := 0; i < b.N; i++ {
+		ko = &pb.BlockOffset{
+			Key: []byte{1, 23},
+		}
+	}
+}
+
+func TestBlockOffsetSizeCalculation(t *testing.T) {
+	// Empty struct testing.
+	require.Equal(t, calculateOffsetsSize([]*pb.BlockOffset{&pb.BlockOffset{}}), int64(88))
+	// Testing with key bytes
+	require.Equal(t, calculateOffsetsSize([]*pb.BlockOffset{&pb.BlockOffset{Key: []byte{1, 1}}}),
+		int64(90))
+}
